//control_distance_sensor.c
//距離センサに関連する処理をおこなうモジュール

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>

#include <pigpio.h>
#include "control_distance_sensor.h"
#include "control_common.h"

#define DISTANCE_SENSOR_INTERVAL_MS (50)

//距離センサが接続されているGPIO番号
#define GPIO_NUM_DIST_TRIG (27)
#define GPIO_NUM_DIST_ECHO (26)

//移動平均フィルタの長さ
#define FILTER_LENGTH_DISTANCE_SENSOR (5)

//パルスが立ち上がったときと立ち下がったときの時間
typedef struct {
    int rising_time;  // 立ち上がり時刻
    int falling_time; // 立ち下がり時刻
} PulseData;

//mutex変数を初期化
static pthread_mutex_t mutex_port_state_raw = PTHREAD_MUTEX_INITIALIZER;

static bool distance_sensor_thread_terminate_request;   //スレッド停止要求
static bool distance_sensor_thread_terminated;          //スレッドが停止したか

//移動平均フィルタの値を保存する配列
static int filter_distance_sensor[FILTER_LENGTH_DISTANCE_SENSOR];

static int distance_cm; //センサ出力をcm単位で表記
static int pulse_width; //パルス幅を表すレジスタ生値

//パルスの立ち上がり/立ち下がり時刻を保存する
static PulseData pulse_data = {0, 0};

//スレッド関数
static void* thread_func_distance_sensor(void* arg);

static void gpio_irq_handler(int gpio, int level, uint32_t tick, void *userdata);

//DISTANCE SENSORを初期化する
void distance_sensor_init(void){
    pthread_t thread_distance_sensor_id;

    //関連するGPIOを初期化する
    gpioSetMode(GPIO_NUM_DIST_TRIG, PI_OUTPUT);  //TRIGは出力とする
    gpioSetMode(GPIO_NUM_DIST_ECHO, PI_INPUT);  //ECHOは入力とす
    gpioWrite(GPIO_NUM_DIST_TRIG, 0);   //TRIG信号をLOWに設定する

    pulse_width = 0;
    distance_cm = 0;

    //フィルタを初期化
    for(int i = 0; i < FILTER_LENGTH_DISTANCE_SENSOR; i++){
        filter_distance_sensor[i] = 0;
    }

     // スレッド生成
    if (pthread_create(&thread_distance_sensor_id, NULL, thread_func_distance_sensor, NULL) != 0) {
        perror("pthread_create");
        exit(-1);
    }
    pthread_detach(thread_distance_sensor_id);
}


// スレッド関数
static void* thread_func_distance_sensor(void* arg) {

    //IRQの割り込みハンドラを登録(pigpioライブラリの機能を利用する)
    gpioSetAlertFuncEx(GPIO_NUM_DIST_ECHO, gpio_irq_handler, &pulse_data);

    while (1) {
         //スレッド終了要求があれば、スレッドを終了する(無限ループを抜ける)
       if(distance_sensor_thread_terminate_request == true){
            break;
        }
        //距離の取得間隔だけ待つ
        sleep_ms(DISTANCE_SENSOR_INTERVAL_MS);

        //トリガ信号を出力する
        gpioWrite(GPIO_NUM_DIST_TRIG, 1);   //トリガ信号をHiにする
        sleep_us(10); //10usの間待つ
        gpioWrite(GPIO_NUM_DIST_TRIG, 0);   //トリガ信号をLoにする
        //この後、ECHO信号が返ってくれば外部割り込みが発生するはずなので、距離は外部割り込みで判定する

    }

    distance_sensor_thread_terminated = true;
    return NULL; // 終了
}

//パルス幅を表す値を取得する
int distance_sensor_get_raw_pulse_width(void){
    return (pulse_width);
}

//距離センサの出力をcm単位で取得する
int distance_sensor_get_distance(void){
    return (distance_cm);
}

// GPIO 26のIRQ割り込みハンドラ
static void gpio_irq_handler(int gpio, int level, uint32_t tick, void *userdata) {
    static int filter_pos = 0;
    long int filter_sum = 0;
    PulseData *data = (PulseData *)userdata;

    //レベルがHiなら立ち上がり
    if (level == 1) { // 立ち上がり
        data->rising_time = tick;
    } else if (level == 0) { // 立ち下がり
        data->falling_time = tick;

        // パルス幅を計算して表示
        int width = data->falling_time - data->rising_time;
        if (width >= 0) {
            //printf("Pulse width: %d microseconds\n", pulse_width);
            pulse_width = width;
        } 

        filter_distance_sensor[filter_pos] = width;
        filter_pos ++;
        if(filter_pos > FILTER_LENGTH_DISTANCE_SENSOR){
            filter_pos = 0;
        }

        for(int i = 0; i < FILTER_LENGTH_DISTANCE_SENSOR; i++){
            filter_sum += filter_distance_sensor[i];
        }
        
        distance_cm = filter_sum / FILTER_LENGTH_DISTANCE_SENSOR / 58;


    }
}


//スレッドを終了させる。スレッドが終了するまでブロックする
void terminate_distance_sensor(void){
    distance_sensor_thread_terminated = false;
    distance_sensor_thread_terminate_request = true;

    printf("distance_sensor thread terminating.\n");
    while(1){
        sleep_ms(10);   //10msごとにスレッドが終了したかチェックする
        if(distance_sensor_thread_terminated == true){
            break;
        }
    }

    printf("distance_sensor thread terminated.\n");

}


