//control_sw.c

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <signal.h>
#include <unistd.h>

#include <pigpio.h>

#include "control_sw.h"
#include "control_common.h"


#define GPIO_SW_P0 (16)
#define GPIO_SW_P1 (17)
#define GPIO_SW_P2 (20)


#define NUM_OF_SW (3)
#define SW_CHATTERING_COUNT_MAX (4)

static unsigned int sw_hi_count[NUM_OF_SW];
static unsigned int sw_low_count[NUM_OF_SW];
static e_sw_state sw_state[NUM_OF_SW];



//mutex変数を初期化
static pthread_mutex_t mutex_sw = PTHREAD_MUTEX_INITIALIZER;

static bool sw_thread_terminate_request;   //スレッド終了要求フラグ
static bool sw_thread_terminated;          //スレッド終了完了フラグ

//スイッチの読み取り処理を行う
static void sw_check(void);

//スレッド関数
static void* thread_func_sw(void* arg);

//このモジュールを初期化する
void sw_init(void){
    pthread_t thread_sw_id;
    int i;

    for(int i = 0; i < NUM_OF_SW; i++){
        sw_hi_count[i] = 0;
        sw_low_count[i] = SW_CHATTERING_COUNT_MAX + 1;
        sw_state[i] = SW_OFF;
     }

    gpioSetMode(GPIO_SW_P0, PI_INPUT);
    gpioSetMode(GPIO_SW_P1, PI_INPUT);
    gpioSetMode(GPIO_SW_P2, PI_INPUT);

    // スレッド生成
    if (pthread_create(&thread_sw_id, NULL, thread_func_sw, NULL) != 0) {
        perror("pthread_create");
        exit(-1);
    }
    //スレッドにデタッチ属性を設定する
    pthread_detach(thread_sw_id);

}

//スレッドを終了する
void terminate_sw(void){
    sw_thread_terminated = false;       //スレッド終了済みフラグをOFFにする
    sw_thread_terminate_request = true; //スレッド終了要求フラグをONにする

    printf("sw thread terminating.\n");
    while(1){
        sleep_ms(10);
        if(sw_thread_terminated == true){
            break;
        }
    }
    printf("sw thread terminated.\n");
}

//スレッド関数
static void* thread_func_sw(void* arg) {

    while (1) {
        //スレッド終了要求があれば、スレッドを終了する(無限ループを抜ける)
        if(sw_thread_terminate_request == true){
            break;
        }
        //1サイクルは10ms
        sleep_ms(10);
        sw_check();

    }

    sw_thread_terminated = true;
    return NULL; // 終了
}


//スイッチの状態をチェックする(thread_func_swスレッドから呼ばれる)
static void sw_check(void){
    if(gpioRead(GPIO_SW_P0) == 0){
        sw_low_count[0] ++;
        sw_hi_count[0] = 0;
    } else if (gpioRead(GPIO_SW_P0) == 1){
        sw_low_count[0] = 0;
        sw_hi_count[0]  ++;

    } else {
        //do nothing
    }
    
    //SWがON入力だったときはONカウントをインクリメント、OFF入力だったらOFFカウントをインクリメントする
    if(gpioRead(GPIO_SW_P1) == 0){
        sw_low_count[1] ++;
        sw_hi_count[1] = 0;
    } else if (gpioRead(GPIO_SW_P1) == 1){
        sw_low_count[1] = 0;
        sw_hi_count[1]  ++;

    } else {
        //do nothing
    }

    if(gpioRead(GPIO_SW_P2) == 0){
        sw_low_count[2] ++;
        sw_hi_count[2] = 0;
    } else if (gpioRead(GPIO_SW_P2) == 1){
        sw_low_count[2] = 0;
        sw_hi_count[2]  ++;

    } else {
        //do nothing
    }

    pthread_mutex_lock(&mutex_sw);    //排他制御のため、mutexロックを取得する

    for(int sw = 0; sw < NUM_OF_SW; sw++){
        if(sw_low_count[sw] == SW_CHATTERING_COUNT_MAX){
            sw_state[sw] = SW_TOGGLED_OFF; //OFFに切り替わる
        } else if(sw_low_count[sw] > SW_CHATTERING_COUNT_MAX){
            sw_state[sw] = SW_OFF; //OFF確定
            sw_low_count[sw] = SW_CHATTERING_COUNT_MAX + 1;
        }

        if(sw_hi_count[sw] == SW_CHATTERING_COUNT_MAX){
            sw_state[sw] = SW_TOGGLED_ON; //ONに切り替わる
        } else if(sw_hi_count[sw] > SW_CHATTERING_COUNT_MAX){
            sw_state[sw] = SW_ON; //ON確定
            sw_hi_count[sw] = SW_CHATTERING_COUNT_MAX + 1;
        } else {
            //do nothing
        }
    }
    pthread_mutex_unlock(&mutex_sw); //mutexロックを解放する
}

//スイッチの状態を取得して返す(他のスレッドから呼ばれる)
e_sw_state sw_get_state(e_sw_num sw_num){
    e_sw_state result;
    
    pthread_mutex_lock(&mutex_sw);    //排他制御のため、mutexロックを取得する

    result = sw_state[sw_num]; 
    
    pthread_mutex_unlock(&mutex_sw); //mutexロックを解放する
    return(result);
}