//vehicle_control_i2c.c
//I2C関連の処理をおこなうモジュール

#include <stdio.h>

#include <pigpio.h>
#include <fcntl.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <linux/i2c-dev.h>

#include <stdlib.h>

#include "control_common.h"

#define I2C_DEV_FILE "/dev/i2c-1"  // Raspberry PiのI2Cデバイスファイル


//mutex変数を初期化
static pthread_mutex_t mutex_i2c = PTHREAD_MUTEX_INITIALIZER;

//I2Cデバイスに1byte書き込む
void i2c_write_byte(unsigned char addr, unsigned char reg_addr, unsigned char data){
    int file;
    char buf[2] = {reg_addr, data};  // 送信するコマンドとデータ

    //mutexロックを獲得
    pthread_mutex_lock(&mutex_i2c);

    // I2Cデバイスを開く
    if ((file = open(I2C_DEV_FILE, O_RDWR)) < 0) {
        perror("Failed to open the i2c bus");
        exit(1);
    }

    // I2Cスレーブデバイスのアドレスを設定
    if (ioctl(file, I2C_SLAVE, addr) < 0) {
        perror("Failed to acquire bus access and/or talk to slave");
        close(file);
        exit(1);
    }

    // データを送信
    if (write(file, buf, 2) != 2) {
        perror("Failed to write to the i2c bus");
        close(file);
        exit(1);
    }

    close(file);

    sleep_us(10);
    //mutexロックを解放
    pthread_mutex_unlock(&mutex_i2c);


}

//I2Cデバイスから1byte読み出す
unsigned char i2c_read_byte(unsigned char addr, unsigned char reg_addr){
    int file;
    char buffer[1];
    
    //mutexロックを獲得
    pthread_mutex_lock(&mutex_i2c);

    // I2Cデバイスファイルを開く
    if ((file = open(I2C_DEV_FILE, O_RDWR)) < 0) {
        perror("Failed to open the I2C bus");
        return -1;
    }

    // I2Cアドレスを設定
    if (ioctl(file, I2C_SLAVE, addr) < 0) {
        perror("Failed to set I2C address");
        close(file);
        return -1;
    }

    // 読み取り対象のレジスタを設定
    buffer[0] = reg_addr;
    if (write(file, buffer, 1) != 1) {
        perror("Failed to write to the I2C device");
        close(file);
        return -1;
    }

    // レジスタから1バイトを読み取る
    if (read(file, buffer, 1) != 1) {
        perror("Failed to read from the I2C device");
        close(file);
        return -1;
    }

    // デバイスファイルを閉じる
    close(file);
    
    sleep_us(10);
    //mutexロックを解放
    pthread_mutex_unlock(&mutex_i2c);

    return(buffer[0]);

}

//I2Cデバイスに複数バイト書き込む
void i2c_write_multi_byte(int addr, const unsigned char *data, int length) 
{
    int file;

    //mutexロックを獲得
    pthread_mutex_lock(&mutex_i2c);


    // I2Cデバイスファイルをオープン
    file = open(I2C_DEV_FILE, O_RDWR);
    if (file < 0) {
        perror("I2Cデバイスファイルのオープンに失敗しました");
        exit(-1);
    }

    // スレーブアドレスを設定
    if (ioctl(file, I2C_SLAVE, addr) < 0) {
        perror("I2Cスレーブアドレスの設定に失敗しました");
        close(file);
        exit(-1);
    }

    // データを送信
    if (write(file, data, length) != length) {
        perror("I2Cデータ送信に失敗しました");
        close(file);
        exit(-1);
    }

    // デバイスファイルをクローズ
    close(file);
    //mutexロックを解放
    sleep_us(10);
    pthread_mutex_unlock(&mutex_i2c);    
    
}
