//vehicle_control_adc.c
//ADコンバータに関連する処理をおこなうモジュール

#include <stdio.h>

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/spi/spidev.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>
#include <time.h>

#include "vehicle_control_adc.h"
#include "vehicle_control_common.h"


#define SPI_DEVICE "/dev/spidev0.1" // SPIデバイスファイル
#define SPI_SPEED 1000000 // SPI通信速度 (1MHz)
#define MCP3208_CHANNELS 8 // MCP3208のチャンネル数

//ADCスレッドが動作する時間間隔(ms)
#define ADC_THREAD_INTERVAL_MS (50)


static bool adc_thread_terminate_request;
static bool adc_thread_terminated;

//AD変換結果を格納する配列
static unsigned int ad_value[MCP3208_CHANNELS];


//mutex変数を初期化
static pthread_mutex_t mutex_ad_val = PTHREAD_MUTEX_INITIALIZER;


static void set_ad_val(unsigned int ad[]);
static void* thread_func_adc(void* arg) ;
static int read_adc(int fd, int adc_channel);


//ADCを初期化する
void adc_init(void){
    pthread_t thread_adc_id;

    // スレッド生成
    if (pthread_create(&thread_adc_id, NULL, thread_func_adc, NULL) != 0) {
        perror("pthread_create");
        exit(-1);
    }
    pthread_detach(thread_adc_id);
}


// スレッド関数
static void* thread_func_adc(void* arg) {
   int fd;

   //AD変換結果を一時的に保存しておくローカル変数
   unsigned int adc_buf[MCP3208_CHANNELS] = {0,0,0,0,0,0,0,0};

    printf("mcp3208 initializing\n");

    for (int channel = 0; channel < MCP3208_CHANNELS; channel++) {
            
        ad_value[channel] =0;

    }
    // SPIデバイスファイルを開く
    fd = open(SPI_DEVICE, O_RDWR);
    if (fd < 0) {
        perror("Failed to open SPI device");
        exit(0);
    }

    // SPI設定
    unsigned int mode = 0;
    unsigned int bits_per_word = 8;
    unsigned int speed = SPI_SPEED;

    if (ioctl(fd, SPI_IOC_WR_MODE, &mode) < 0 || ioctl(fd, SPI_IOC_RD_MODE, &mode) < 0) {
        perror("Failed to set SPI mode");
        close(fd);
        exit(-1);
    }

    if (ioctl(fd, SPI_IOC_WR_BITS_PER_WORD, &bits_per_word) < 0 || ioctl(fd, SPI_IOC_RD_BITS_PER_WORD, &bits_per_word) < 0) {
        perror("Failed to set bits per word");
        close(fd);
        exit(-1);
    }

    if (ioctl(fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed) < 0 || ioctl(fd, SPI_IOC_RD_MAX_SPEED_HZ, &speed) < 0) {
        perror("Failed to set SPI speed");
        close(fd);
        exit(-1);
    }

    
    while (1) {
        //スレッド終了要求があれば、スレッドを終了する(無限ループを抜ける)
        if(adc_thread_terminate_request == true){
            break;
        }
        sleep_ms(ADC_THREAD_INTERVAL_MS);
        // 各チャンネルの値を取得
        for (int channel = 0; channel < MCP3208_CHANNELS; channel++) {
            
            adc_buf[channel] = read_adc(fd, channel);
        }
        set_ad_val(adc_buf);
    }
    // デバイスファイルを閉じる
    close(fd);

    adc_thread_terminated = true;
    return NULL; //terminate thread
}

//取得したAD値をモジュール変数に保存する
static void set_ad_val(unsigned int ad[]){
    pthread_mutex_lock(&mutex_ad_val);
    for (int channel = 0; channel < MCP3208_CHANNELS; channel++) {
            
        ad_value[channel] = ad[channel];

    }

    pthread_mutex_unlock(&mutex_ad_val);
}

//AD変換された値を取得する
unsigned int adc_get_value(int channel){
    unsigned int val = 0;
    if((channel < 0) || (channel >= MCP3208_CHANNELS) ){

    }

    pthread_mutex_lock(&mutex_ad_val);
    val = ad_value[channel];
    pthread_mutex_unlock(&mutex_ad_val);

    return val;
}

//スレッドを終了させる。スレッドが終了するまでブロックする
void terminate_adc(void){
    adc_thread_terminated = false;
    adc_thread_terminate_request = true;

    printf("adc thread terminating.\n");
    while(1){
        sleep_10ms; //sleep 10ms
        if(adc_thread_terminated == true){
            break;
        }
    }

    printf("adc thread terminated.\n");

}


// MCP3208から1つのチャンネルの値を取得する関数
static int read_adc(int fd, int adc_channel) {
    if (adc_channel < 0 || adc_channel >= MCP3208_CHANNELS) {
        fprintf(stderr, "Invalid ADC channel: %d\n", adc_channel);
        return -1;
    }

    // MCP3208への送信データ (3バイト)
    unsigned char tx_buffer[3];
    unsigned char rx_buffer[3];
    struct spi_ioc_transfer spi_transfer = {
        .tx_buf = (unsigned long)tx_buffer,
        .rx_buf = (unsigned long)rx_buffer,
        .len = 3,
        .speed_hz = SPI_SPEED,
        .bits_per_word = 8,
    };

    tx_buffer[0] = 0b00000110 | ((adc_channel & 0b100) >> 2); // スタートビットとシングルエンド設定
    tx_buffer[1] = (adc_channel & 0b011) << 6; // チャンネル選択
    tx_buffer[2] = 0; // ダミーデータ

    // SPI通信
    if (ioctl(fd, SPI_IOC_MESSAGE(1), &spi_transfer) < 0) {
        perror("SPI communication failed");
        return -1;
    }

    // ADC値の計算 (12ビットの結果を取得)
    int result = ((rx_buffer[1] & 0x0F) << 8) | rx_buffer[2];
    return result;
}

