//vehicle_control_communication.c
//通信に関連する処理をおこなうモジュール

#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdbool.h>
#include <fcntl.h>
#include "vehicle_control_communication.h"
#include "vehicle_control_common.h"

// static void communication_close_sock_fd(void);

#define UDP_PORT 12345 // 待ち受けポート番号
#define BUFFER_SIZE 1024

//UDP通信に使うソケットディスクリプタ
static int sockfd;

//mutex変数を初期化
static pthread_mutex_t mutex_new_packet = PTHREAD_MUTEX_INITIALIZER;

//受信データを保存するバッファ
static unsigned char recv_payload[UDP_PAYLOAD_LENGTH_MAX];

//新しいパケットが届いたか
static bool new_data_exist;

static bool terminate_communication_flag;
static bool communication_thread_terminated;

//UDP受信するスレッド関数
static void *thread_func_communication(void *arg);
static     pthread_t communication_thread_id;

//このモジュールを初期化する
void communication_init(void)
{
    struct sockaddr_in server_addr;

    terminate_communication_flag = false;

    // パケット受信スレッドを作成
    if (pthread_create(&communication_thread_id, NULL, thread_func_communication, NULL) != 0) {
        perror("pthread_create failed");
        close(sockfd);
        exit(EXIT_FAILURE);
    }

    // スレッドをデタッチ属性にして、終了を待ち受けない
    pthread_detach(communication_thread_id);

    return ;


}

//スレッドを終了させる。スレッドが終了するまでブロックする
void terminate_communication(void){
    communication_thread_terminated = false;

    //スレッド終了要求フラグをONにする
    terminate_communication_flag = true;

    printf("comm thread terminating.\n");
    //スレッドが終了されるのを待つ
    while(1){
        sleep_10ms();   //10msごとにスレッドが終了したかチェックする
        if(communication_thread_terminated == true){
            break;
        }
    }

    printf("comm thread terminated.\n");

}


//新たなパケットデータが来たので、それを反映させる
static void communication_set_new_data(char *data){
    //mutexをロック
    pthread_mutex_lock(&mutex_new_packet);
    strncpy(recv_payload, data, UDP_PAYLOAD_LENGTH_MAX);

    new_data_exist = true;
    //mutexをアンロック
    pthread_mutex_unlock(&mutex_new_packet);
}

//新たなパケットがあるか
bool is_new_packet_exists(void){
    bool data;
    //printf("is_new_packet_exists\n");
    //mutexをロック
    pthread_mutex_lock(&mutex_new_packet);
    data = new_data_exist;
    //mutexをアンロック
    pthread_mutex_unlock(&mutex_new_packet);

    return(data);
}

//直近に到着したパケットを取得する
void communication_get_new_data(char *data){
    //mutexをロック
    pthread_mutex_lock(&mutex_new_packet);
    strncpy(data, recv_payload,  UDP_PAYLOAD_LENGTH_MAX);

    new_data_exist = false;
    //mutexをアンロック
    pthread_mutex_unlock(&mutex_new_packet);

}

//スレッド関数
static void *thread_func_communication(void *arg) {
    int sockfd; //ソケットディスクリプタ
    struct sockaddr_in servaddr, cliaddr;
    char buffer[BUFFER_SIZE];   //受信データを保存しておくバッファ
    socklen_t len;
    fd_set readfds; //ファイルディスクリプタセット
    int n;

    // ソケットを作成
    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        perror("socket creation failed");
        exit(EXIT_FAILURE);
    }

    memset(&servaddr, 0, sizeof(servaddr));
    memset(&cliaddr, 0, sizeof(cliaddr));

    // サーバアドレスの設定
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = INADDR_ANY;
    servaddr.sin_port = htons(UDP_PORT);

    // ソケットにアドレスをバインド
    if (bind(sockfd, (const struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
        perror("bind failed");
        exit(EXIT_FAILURE);
    }

    // ソケットをノンブロッキングモードに設定
    //通常の方法だとパケットを受信するまでブロックするが、
    //この方法はパケットが来なくてもブロックせずに次の命令を実行できる
    int flags = fcntl(sockfd, F_GETFL, 0);
    fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);

    // select()のタイムアウト設定（10ms）
    //タイムアウト時間が経過するまでパケットが来なかったら、次の命令が実行さえっる
    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 10000; //10,000 us => 10ms

    // パケットを受信する
    while (1) {
        //スレッド終了要求があれば、スレッドを終了する(無限ループを抜ける)
        if(terminate_communication_flag == true){
            break;
        }

        FD_ZERO(&readfds);
        //使用するソケットディスクリプタを設定
        FD_SET(sockfd, &readfds);

        // パケットが到着しているかをチェック
        int ret = select(sockfd + 1, &readfds, NULL, NULL, &tv);
        if (ret == -1) {
            perror("select error");
            exit(EXIT_FAILURE);
        } else if (ret == 0) {
            // タイムアウト、パケットは到着していない
            continue;
        }

        // ソケットディスクリプタが利用可能かチェック
        if (FD_ISSET(sockfd, &readfds)) {
            //パケットが到着しており、ソケットディスクリプタが利用可能なら、受信して取り込む
            len = sizeof(cliaddr);
            n = recvfrom(sockfd, buffer, BUFFER_SIZE, 0, (struct sockaddr *)&cliaddr, &len);
            if (n < 0) {
                perror("recvfrom error");
                continue;
            }
            buffer[n] = '\0'; // ヌル終端を確実にする
            communication_set_new_data(buffer);

        }
    }

    communication_thread_terminated = true;
    close(sockfd);
    return NULL;
}

