//write_pcal9555apw.c
//GPIO入力に関連する処理をおこなうモジュール

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>

#include <pigpio.h>
#include "vehicle_control_gpio_in.h"
#include "vehicle_control_i2c.h"
#include "vehicle_control_common.h"

#define GPIO_SW_P0 (16)
#define GPIO_SW_P1 (17)
#define GPIO_SW_P2 (20)

#define GPIO_RS_DIGITAL_OUT_L0 (22)
#define GPIO_RS_DIGITAL_OUT_L1 (23)
#define GPIO_RS_DIGITAL_OUT_R0 (24)
#define GPIO_RS_DIGITAL_OUT_R1 (25)

//PCAL9555APWのI2Cスレーブアドレス
#define ADDR_PCAL9555APW (0x20)

//
#define SW_CHATTERING_COUNT_MAX (3)

//
static void set_gpio_in_raw_sate(int *state);
static void set_sw_state(e_sw_state *state);
static void init_PCAL9555APW(void);     //PCAL9555APWを初期化する
static void* thread_func_gpio_in(void* arg);

//mutex変数を初期化
static pthread_mutex_t mutex_port_state_raw = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t mutex_gpio_sw_state = PTHREAD_MUTEX_INITIALIZER;


static bool gpio_in_thread_terminate_request;   //スレッド終了要求フラグ
static bool gpio_in_thread_terminated;          //スレッドが終了したか

//GPIOのポート番号を配列で保存。列挙型 "e_gpio_input" の番号と同じ添字の番号になる
static int gpio_port_nums[] = {
    GPIO_SW_P0, GPIO_SW_P1, GPIO_SW_P2,
    GPIO_RS_DIGITAL_OUT_L0, GPIO_RS_DIGITAL_OUT_L1,
    GPIO_RS_DIGITAL_OUT_R0, GPIO_RS_DIGITAL_OUT_R1
};

//取得したポートの状態を保存しておくバッファ
static int gpio_in_raw_sate[RAW_GPIO_GPIO_INPUT_MAX];

static unsigned int sw_on_count[RAW_GPIO_GPIO_INPUT_MAX];
static unsigned int sw_off_count[RAW_GPIO_GPIO_INPUT_MAX];
static e_sw_state   gpio_in_sw_state[RAW_GPIO_GPIO_INPUT_MAX];


//GPIO_INを初期化する
//※LEDの初期化が終わってから呼ぶこと
void gpio_in_init(void){
    pthread_t thread_gpio_in_id;

    for(e_gpio_input port=0; port < (sizeof(gpio_port_nums)/sizeof(int)); port ++){
        gpioSetMode(gpio_port_nums[port], PI_INPUT);
    }

    for (e_sw_input sw = SW_INPUT_START; sw < SW_INPUT_MAX; sw++){
            gpio_in_sw_state[sw] = SW_OFF;
            sw_on_count[sw] = 0;
            sw_off_count[sw] = 0;
     }
    init_PCAL9555APW();

     // スレッド生成
    if (pthread_create(&thread_gpio_in_id, NULL, thread_func_gpio_in, NULL) != 0) {
        perror("pthread_create");
        exit(-1);
    }


}

//PCAL9555APWを初期化する(LEDの初期化後に呼ぶこと)
static void init_PCAL9555APW(void){
    //Port1のb0, b1を入力にする
    i2c_write_byte(ADDR_PCAL9555APW, 0x07, 0b00000011);

}


e_sw_state gpio_in_get_sw_state(e_sw_input sw){
    e_sw_state result;
    
    pthread_mutex_lock(&mutex_gpio_sw_state);    //mutexロックを取得する

    result = gpio_in_sw_state[sw]; 
    
    pthread_mutex_unlock(&mutex_gpio_sw_state); //mutexロックを解放する
    return(result);
}


//GPIOの状態を取得する
int gpio_in_get_state_raw(e_gpio_input port){
    int state = 0;

    //mutexロックを取得する
    pthread_mutex_lock(&mutex_port_state_raw);
    if((port < RAW_GPIO_INPUT_START) ||(port >= RAW_GPIO_GPIO_INPUT_MAX)){
        state = 0;
    } else {
        state = gpio_in_raw_sate[port];
    }
    //mutexロックを解放する
    pthread_mutex_unlock(&mutex_port_state_raw);
    return(state);
}

//スレッドを終了する
void terminate_gpio_in(void){
    gpio_in_thread_terminated = false;
    gpio_in_thread_terminate_request = true;

    printf("gpio_in thread terminating.\n");
    while(1){
        sleep_10ms();//10msごとにスレッドが終了したかチェックする
        if(gpio_in_thread_terminated == true){
            break;
        }
    }

    printf("gpio_in thread terminated.\n");

}


// スレッド関数
static void* thread_func_gpio_in(void* arg) {
    int port_state[RAW_GPIO_GPIO_INPUT_MAX];
    e_sw_state sw_state[SW_INPUT_MAX];
    unsigned char pcal_data;

    for (e_sw_input sw = SW_INPUT_START; sw < SW_INPUT_MAX; sw++){
            sw_state[sw] = SW_OFF;
    }

    while (1) {
        //スレッド終了要求があれば、スレッドを終了する(無限ループを抜ける)
        if(gpio_in_thread_terminate_request == true){
            break;
        }
        //1サイクルは10ms
        sleep_10ms();
        
        //ポート状態を読み出す
        for(e_gpio_input port=0; port < (sizeof(gpio_port_nums)/sizeof(int)); port ++){
            port_state[port] = gpioRead(gpio_port_nums[port]);
        }

        //読み出したポート状態をモジュール変数に保存する
        set_gpio_in_raw_sate(port_state);   

        //PCAL9555APWから読み出す
        pcal_data = i2c_read_byte(ADDR_PCAL9555APW, 0x01);
        port_state[SW_T0] = pcal_data & 0b00000001;
        port_state[SW_T1] = (pcal_data & 0b00000010) >> 1;

        //各ポートをスイッチとして取り扱う場合の、チャタリング処理
        //SWがON入力だったときはONカウンタをインクリメント、OFF入力だったらOFFカウンタをインクリメントする
        sw_on_count[SW_P0]  += ( (port_state[RAW_GPIO_SW_P0] == 1) ? 1 : 0 ); //Hi: on, Low: off
        sw_off_count[SW_P0] += ( (port_state[RAW_GPIO_SW_P0] == 0) ? 1 : 0 );
        sw_on_count[SW_P1]  += ( (port_state[RAW_GPIO_SW_P1] == 1) ? 1 : 0 );
        sw_off_count[SW_P1] += ( (port_state[RAW_GPIO_SW_P1] == 0) ? 1 : 0 );
        sw_on_count[SW_P2]  += ( (port_state[RAW_GPIO_SW_P2] == 1) ? 1 : 0 );
        sw_off_count[SW_P2] += ( (port_state[RAW_GPIO_SW_P2] == 0) ? 1 : 0 );

        sw_on_count[SW_T0]  += ( (port_state[RAW_GPIO_SW_T0] == 0) ? 1 : 0 ); //Low: on, Hi: off
        sw_off_count[SW_T0] += ( (port_state[RAW_GPIO_SW_T0] == 1) ? 1 : 0 );
        sw_on_count[SW_T1]  += ( (port_state[RAW_GPIO_SW_T1] == 0) ? 1 : 0 ); 
        sw_off_count[SW_T1] += ( (port_state[RAW_GPIO_SW_T1] == 1) ? 1 : 0 );

        sw_on_count[RS_DIGITAL_OUT_L0]  += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_L0] == 0) ? 1 : 0 ); 
        sw_off_count[RS_DIGITAL_OUT_L0] += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_L0] == 1) ? 1 : 0 );
        sw_on_count[RS_DIGITAL_OUT_L1]  += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_L1] == 0) ? 1 : 0 ); 
        sw_off_count[RS_DIGITAL_OUT_L1] += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_L1] == 1) ? 1 : 0 );
        sw_on_count[RS_DIGITAL_OUT_R0]  += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_R0] == 0) ? 1 : 0 ); 
        sw_off_count[RS_DIGITAL_OUT_R0] += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_R0] == 1) ? 1 : 0 );
        sw_on_count[RS_DIGITAL_OUT_R1]  += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_R1] == 0) ? 1 : 0 ); 
        sw_off_count[RS_DIGITAL_OUT_R1] += ( (port_state[RAW_GPIO_RS_DIGITAL_OUT_R1] == 1) ? 1 : 0 );

        //ONカウンタ、OFFカウンタが指定のカウント以上なら、その状態を確定する
        for(int sw = 0; sw < SW_INPUT_MAX; sw++){
           if(sw_off_count[sw] == SW_CHATTERING_COUNT_MAX){
                sw_state[sw] = SW_TOGGLED_OFF; //OFFに切り替わる 
                sw_on_count[sw] = 0;
            } else if(sw_off_count[sw] > SW_CHATTERING_COUNT_MAX){
                sw_state[sw] = SW_OFF; //OFF確定
                sw_off_count[sw] = SW_CHATTERING_COUNT_MAX + 1;
            }

            if(sw_on_count[sw] == SW_CHATTERING_COUNT_MAX){
                sw_state[sw] = SW_TOGGLED_ON; //ONに切り替わる
                sw_off_count[sw] = 0;          
            } else if(sw_on_count[sw] > SW_CHATTERING_COUNT_MAX){
                sw_state[sw] = SW_ON; //ON確定
                sw_on_count[sw] = SW_CHATTERING_COUNT_MAX + 1;
            } else {
                //do nothing
            }
        }

        //スイッチの状態とポートの状態を保存する
        set_sw_state(sw_state);
        set_gpio_in_raw_sate(port_state);   

    }

    gpio_in_thread_terminated = true;
    return NULL; // 終了
}

//取得したポート状態をバッファに保存する
static void set_gpio_in_raw_sate(int *state){
   pthread_mutex_lock(&mutex_port_state_raw);
    for(e_gpio_input port=RAW_GPIO_INPUT_START; port < RAW_GPIO_GPIO_INPUT_MAX; port ++){
        gpio_in_raw_sate[port] = state[port];
    }

   pthread_mutex_unlock(&mutex_port_state_raw);

}

//取得したスイッチの状態をバッファに保存する
static void set_sw_state(e_sw_state *state){
   pthread_mutex_lock(&mutex_gpio_sw_state);
    for(e_sw_input sw=SW_INPUT_START; sw < SW_INPUT_MAX; sw ++){
        gpio_in_sw_state[sw] = state[sw];
    }

   pthread_mutex_unlock(&mutex_gpio_sw_state);

}
