//vehicle_control_led.c
//LEDに関連する処理をおこなうモジュール

#include <stdio.h>
#include <pigpio.h>
#include <fcntl.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <linux/i2c-dev.h>

#include <stdlib.h>

#include "vehicle_control_led.h"
#include "vehicle_control_i2c.h"

//GPIOで点灯するLEDとGPIO番号の対応
#define GPIO_LED_ST0 4
#define GPIO_LED_ST1 5
#define GPIO_LED_ST2 6

//LED_IND0〜LED_IND7のLEDビット位置
#define PCAL_LED_IND0 0b00000001
#define PCAL_LED_IND1 0b00000010
#define PCAL_LED_IND2 0b00000100
#define PCAL_LED_IND3 0b00001000
#define PCAL_LED_IND4 0b00010000
#define PCAL_LED_IND5 0b00100000
#define PCAL_LED_IND6 0b01000000
#define PCAL_LED_IND7 0b10000000

//PCAL9555APWのI2Cアドレス
#define ADDR_PCAL9555APW (0x20)

//PCAL9555APW(I2Cデバイス)の初期化
static void init_PCAL9555APW(void);

//LED_IND0〜LED_IND7の点灯パターンを保存する
static unsigned char bit_pattern_INDx;

//このモジュールの初期化
void led_init(void){
    //RaspberryPiのGPIOに接続されているLEDの初期化
    gpioSetMode(GPIO_LED_ST0, PI_OUTPUT);
    gpioSetMode(GPIO_LED_ST1, PI_OUTPUT);
    gpioSetMode(GPIO_LED_ST2, PI_OUTPUT);

    //INDxの最初のビットパターンを設定
    bit_pattern_INDx = 0b00000000;
    //PCAL9555APWを初期化する
    init_PCAL9555APW();

}

//PCAL9555APW(I2Cデバイス)の初期化
static void init_PCAL9555APW(void){
    //port0をすべて出力(0)
    i2c_write_byte(ADDR_PCAL9555APW, 0x06, 0x00);

    //最初の出力パターンをセット
    i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
}

//LEDを指定して点灯させる
void led_turn_on(e_led led){
    switch(led){
    case LED_ST0:
        gpioWrite(GPIO_LED_ST0, 0); /*Low:点灯, Hi: 消灯*/
        break;
    case LED_ST1:
        gpioWrite(GPIO_LED_ST1, 0);
        break;
    case LED_ST2:
        gpioWrite(GPIO_LED_ST2, 0);
        break;
    case LED_IND0:
        bit_pattern_INDx |= PCAL_LED_IND0;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND1:
        bit_pattern_INDx |= PCAL_LED_IND1;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND2:
        bit_pattern_INDx |= PCAL_LED_IND2;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND3:
        bit_pattern_INDx |= PCAL_LED_IND3;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND4:
        bit_pattern_INDx |= PCAL_LED_IND4;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND5:
        bit_pattern_INDx |= PCAL_LED_IND5;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND6:
        bit_pattern_INDx |= PCAL_LED_IND6;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND7:
        bit_pattern_INDx |= PCAL_LED_IND7;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
   }
}

//LEDを指定して消灯させる
void led_turn_off(e_led led){
    switch(led){
    case LED_ST0:
        gpioWrite(GPIO_LED_ST0, 1); /*Low:点灯, Hi: 消灯*/
        break;
    case LED_ST1:
        gpioWrite(GPIO_LED_ST1, 1);
        break;
    case LED_ST2:
        gpioWrite(GPIO_LED_ST2, 1);
        break;
   case LED_IND0:
        bit_pattern_INDx &= ~PCAL_LED_IND0;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND1:
        bit_pattern_INDx &= ~PCAL_LED_IND1;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND2:
        bit_pattern_INDx &= ~PCAL_LED_IND2;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND3:
        bit_pattern_INDx &= ~PCAL_LED_IND3;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND4:
        bit_pattern_INDx &= ~PCAL_LED_IND4;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND5:
        bit_pattern_INDx &= ~PCAL_LED_IND5;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND6:
        bit_pattern_INDx &= ~PCAL_LED_IND6;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    case LED_IND7:
        bit_pattern_INDx &= ~PCAL_LED_IND7;
        i2c_write_byte(ADDR_PCAL9555APW, 0x02, bit_pattern_INDx);
        break;
    }
}

//LEDをすべて消灯する
void led_all_off(void){
    led_turn_off(LED_ST0);
    led_turn_off(LED_ST1);
    led_turn_off(LED_ST2);
}
