//車両の制御をおこなう
//vehicle_control_main.c

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <signal.h>
#include <unistd.h>
#include <pigpio.h>

#include "vehicle_control_main.h"
#include "vehicle_control_led.h"
#include "vehicle_control_gpio_in.h"
#include "vehicle_control_adc.h"
#include "vehicle_control_communication.h"
#include "vehicle_control_motor.h"
#include "vehicle_control_buzzer.h"
#include "vehicle_control_distance_sensor.h"
#include "vehicle_control_lcd.h"
#include "vehicle_control_common.h"

//UDPパケットが来ないのが何サイクルつづくと、データをOFFとするか
#define NO_DATA_COUNT_MAX (50)

//プログラムが起動してからの時間カウント
unsigned long int uptime;

//メイン処理の1サイクルあたりの時間(ms単位)
#define MAIN_INTERVAL_MS (50)

//UDPパケットを受信したときに、LEDを"チカッ"とさせるサイクル数
#define CYCLE_LED_ON_SUSTAIN (10)

//パケット受信が無いときの制御値データの初期値
s_packet_control packet_no_data = {0, 0, 0, 0, 0 ,0};

//sigaction構造体用変数
static struct sigaction act, oldact;

//UDP受信したパケットから制御値を取り出す
static s_packet_control get_control_by_packet(char *payload);

//シグナルハンドラ関数(sigint)
static void sigint_handler(int signum);

int main(void)
{
    //UDPパケットから受信した生データ(payload)
    char udp_packet_payload[UDP_PAYLOAD_LENGTH_MAX];
    unsigned int no_data_count; //何サイクルの間、UDPパケットが来なかったかをカウント 
    unsigned int led_on_sustain_count; //パケットを受信したときにLED ONする継続時間カウント
    char lcd_string[2][LCD_MSG_LENGTH]; //LCDに表示する文字列

    s_packet_control packet, packet_before;    //UDPパケットから取り出した制御値の現在値と、前回値
    char time_now[TIME_STR_MAX]; //現在時刻を格納する

    //パケットから取り出すデータの初期値(ゼロ値)
    packet = packet_no_data;

    //pigpio初期化
    if(gpioInitialise() < 0){
        printf("pigpio initialize error. \n");
        exit(-1);
    }

    printf("vehicle_control started!\n");

    uptime = 0;

    //初期化関数を呼び出し
    led_init();     //LED関連の初期化
    gpio_in_init(); //GPIO入力関連の初期化(LED初期化の後に呼ぶこと)
    adc_init();
    motor_init();
    buzzer_init();
    distance_sensor_init();
    communication_init();
    lcd_init();

	//Ctrl+Cを押した時にシグナルハンドラを実行する
	act.sa_handler = sigint_handler;
	//SIGINT 受信時に動作する
	sigaction(SIGINT, &act, &oldact);

    //UDP受信開始
    printf("vehicle_control started!\n");

    no_data_count = 0;
    led_on_sustain_count = 0;
    packet_before = packet = packet_no_data;
    while(1){
        sleep_ms(MAIN_INTERVAL_MS);
        
        //新たにUDPパケットを受信したら、そのパケットから制御値を取り出し反映させる
        if(is_new_packet_exists() == true){
            //UDPパケットからのデータを取得する
            communication_get_new_data(udp_packet_payload);
            //パケットの内容を解釈して制御値を得る
            packet = get_control_by_packet(udp_packet_payload);

            no_data_count = 0; //データが来たのでカウンタをクリア
            led_turn_off(LED_ST2); //ERROR LEDを消灯
            led_turn_on(LED_ST0);  //STANDBY LEDを点灯
            led_turn_on(LED_ST1);  //RUNNING LEDを点灯

        }

        //LEDを"チカッ"とさせる
        if(led_on_sustain_count > CYCLE_LED_ON_SUSTAIN){
            led_turn_off(LED_ST1);

            led_on_sustain_count = 0;
        }
        //UDPデータが一定時間こなかったら、【全部停止】としてデータを扱う
        if(no_data_count > NO_DATA_COUNT_MAX){
            packet = packet_no_data; //受信データを「全部停止」で上書きする
            led_turn_on(LED_ST2);  //ERROR LEDを点灯する
        }

        //モーターの回転速度を設定する(-100〜0〜100)
        motor_set_speed_left(packet.speed_left_motor);
        motor_set_speed_right(packet.speed_right_motor);

        //ブザーを鳴らす
        if(packet.sw_push0 == 1 ){
            buzzer_set_frequency(440);            
            buzzer_continue();

        } 

        //ブザーを鳴らす（高い音）
        if(packet.sw_push1 == 1 ){
            buzzer_set_frequency(880);            
            buzzer_continue();

        }

        //前部LEDを点灯・消灯する
        if(packet.sw_toggle0 != packet_before.sw_toggle0){
            if(packet.sw_toggle0 == 1){
                led_turn_on(LED_IND0);
                led_turn_on(LED_IND1);
            } else {
                led_turn_off(LED_IND0);
                led_turn_off(LED_IND1);
            }
        }

        //後部LEDを点灯・消灯する
        if(packet.sw_toggle1 != packet_before.sw_toggle1){
            if(packet.sw_toggle1 == 1){
                led_turn_on(LED_IND2);
                led_turn_on(LED_IND3);
                led_turn_on(LED_IND4);
                led_turn_on(LED_IND5);
                led_turn_on(LED_IND6);
                led_turn_on(LED_IND7);
            } else {
                led_turn_off(LED_IND2);
                led_turn_off(LED_IND3);
                led_turn_off(LED_IND4);
                led_turn_off(LED_IND5);
                led_turn_off(LED_IND6);
                led_turn_off(LED_IND7);
            }

        }

        // LCDパネルに
        if((uptime % 5) == 0){
            //LCDパネルに表示する文字列をつくる
            // line 0
            snprintf(lcd_string[0], LCD_MSG_LENGTH, "L%4d%%R%4d%%%1d%1d%1d%1d",
                packet.speed_left_motor, 
                packet.speed_right_motor,
                packet.sw_push0,
                packet.sw_push1,
                packet.sw_toggle0,
                packet.sw_toggle1 );

             //line 1
            snprintf(lcd_string[1], LCD_MSG_LENGTH, "%1d%1d%1d%1d%1d,%1d%1d%1d%1d,%3dcm",
                gpio_in_get_sw_state(RAW_GPIO_SW_P0),
                gpio_in_get_sw_state(RAW_GPIO_SW_P1),
                gpio_in_get_sw_state(RAW_GPIO_SW_P2),
                gpio_in_get_sw_state(RAW_GPIO_SW_T0),
                gpio_in_get_sw_state(RAW_GPIO_SW_T1),            
                gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_L0),
                gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_L1),
                gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_R0),
                gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_R1),
                distance_sensor_get_distance()
            );

            lcd_clear();
            lcd_set_string(LCD_LINE_0, lcd_string[0]);
            lcd_set_string(LCD_LINE_1, lcd_string[1]);
 
        }
 
        //受信データと各部センサおよびAD値、GPIOの読みなどを標準出力に出力する
        get_time_now(time_now);
        printf("%s, %4d, %4d, %2d, %2d, %2d, %2d, %4d, %4d, %4d, %4d, %4d, %4d, %4d, %4d, %2d, %2d, %2d, %2d, %2d, %2d, %2d, %2d, %2d, %6d, %6d\n", time_now,
            packet.speed_left_motor,
            packet.speed_right_motor,
            packet.sw_push0,
            packet.sw_push1,
            packet.sw_toggle0,
            packet.sw_toggle1,
            adc_get_value(0),
            adc_get_value(1),
            adc_get_value(2),
            adc_get_value(3),
            adc_get_value(4),
            adc_get_value(5),
            adc_get_value(6),
            adc_get_value(7),
            gpio_in_get_sw_state(RAW_GPIO_SW_P0),
            gpio_in_get_sw_state(RAW_GPIO_SW_P1),
            gpio_in_get_sw_state(RAW_GPIO_SW_P2),
            gpio_in_get_sw_state(RAW_GPIO_SW_T0),
            gpio_in_get_sw_state(RAW_GPIO_SW_T1),
            gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_L0),
            gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_L1),
            gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_R0),
            gpio_in_get_sw_state(RAW_GPIO_RS_DIGITAL_OUT_R1),
            distance_sensor_get_raw_pulse_width(),
            distance_sensor_get_distance()
        
        );


        //各種カウンタをカウントアップ
        uptime ++; //起動継続時間をカウントアップ
        no_data_count ++;  //データが来なかった時間をカウントアップ
        led_on_sustain_count ++;  //LED ON継続時間をカウントアップ

        packet_before = packet; //現在の値を次のサイクルのために【前回値】として保存する
    }


}

//UDP受信したパケットから制御値を取り出す
static s_packet_control get_control_by_packet(char *payload){
    s_packet_control packet = {};
    int suffix;

    sscanf(payload + sizeof(REMOTE_PACKET_PREFIX), "%d, %d, %d, %d, %d, %d, %d",
        &packet.speed_left_motor,
        &packet.speed_right_motor,
        &packet.sw_push0,
        &packet.sw_push1,
        &packet.sw_toggle0,
        &packet.sw_toggle1,
        &suffix
        
        );

    if( suffix != REMOTE_PACKET_SUFFIX){
        packet = packet_no_data;
    }
    

    return(packet);
}

//シグナルハンドラ関数(sigint)
static void sigint_handler(int signum)
{
    //すべてのスレッドを終了し、このプログラムを終了させる
    terminate_all();
}
